@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2017 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -36,12 +36,8 @@
 #ifndef _LOGGER_H
 #define _LOGGER_H
 
-#ifdef MACOSX
- #ifndef wcscasecmp
-extern inline int wcscasecmp(const wchar_t* s1, const wchar_t* s2);
-  #define MACOSX_ECSCASECMP
- #endif
-#endif
+/* If defined, output debug information about console output. */
+/*#define DEBUG_CONSOLE_OUTPUT*/
 
 #ifdef _DEBUG
  #define _DEBUG_QUEUE
@@ -51,10 +47,10 @@ extern inline int wcscasecmp(const wchar_t* s1, const wchar_t* s2);
  #include <windows.h>
  #define LOG_USER    (1<<3)
 #endif
-#include "wrapper_i18n.h"
 #ifndef DWORD
  #define DWORD unsigned long
 #endif
+#include "logger_base.h"
 
 /* * * Log source constants * * */
 
@@ -87,38 +83,6 @@ extern inline int wcscasecmp(const wchar_t* s1, const wchar_t* s2);
 #define LOG_FORMAT_LOGFILE_DEFAULT TEXT("LPTM")
 #define LOG_FORMAT_CONSOLE_DEFAULT TEXT("PM")
 
-/* * * Log level constants * * */
-
-/* No logging at all. */
-#define LEVEL_NONE   9
-
-/* Notice messages which should always be displayed.  These never go to the syslog. */
-#define LEVEL_NOTICE 8
-
-/* Advisor messages which should always be displayed.  These never go to the syslog. */
-#define LEVEL_ADVICE 7
-
-/* Too many restarts, unable to start etc. Case when the Wrapper is forced to exit. */
-#define LEVEL_FATAL  6
-
-/* JVM died, hung messages */
-#define LEVEL_ERROR  5
-
-/* Warning messages. */
-#define LEVEL_WARN   4
-
-/* Started, Stopped, Restarted messages. */
-#define LEVEL_STATUS 3
-
-/* Copyright message. and logged console output. */
-#define LEVEL_INFO   2
-
-/* Current debug messages */
-#define LEVEL_DEBUG  1
-
-/* Unknown level */
-#define LEVEL_UNKNOWN  0
-
 /* * * Log file roll mode constants * * */
 #define ROLL_MODE_UNKNOWN         0
 #define ROLL_MODE_NONE            1
@@ -132,13 +96,20 @@ extern inline int wcscasecmp(const wchar_t* s1, const wchar_t* s2);
 #define ROLL_MODE_DATE_TOKEN      TEXT("YYYYMMDD")
 
 
-#ifdef WIN32
-extern void setConsoleStdoutHandle(HANDLE stdoutHandle);
-#endif
-
-/* * * Function predeclaration * * */
-#define strcmpIgnoreCase(str1, str2) _tcsicmp(str1, str2)
-
+/* Any log messages generated within signal handlers must be stored until we
+ *  have left the signal handler to avoid deadlocks in the logging code.
+ *  Messages are stored in a round robin buffer of log messages until
+ *  maintainLogger is next called.
+ * When we are inside of a signal, and thus when calling log_printf_queue,
+ *  we know that it is safe to modify the queue as needed.  But it is possible
+ *  that a signal could be fired while we are in maintainLogger, so case is
+ *  taken to make sure that volatile changes are only made in log_printf_queue.
+ */
+#define QUEUE_SIZE 20
+/* The size of QUEUED_BUFFER_SIZE_USABLE is arbitrary as the largest size which can be logged in full,
+ *  but to avoid crashes due to a bug in the HPUX libc (version < 1403), the length of the buffer passed to _vsntprintf must have a length of 1 + N, where N is a multiple of 8. */
+#define QUEUED_BUFFER_SIZE_USABLE (512 + 1)
+#define QUEUED_BUFFER_SIZE (QUEUED_BUFFER_SIZE_USABLE + 4)
 
 /* This can be called from within logging code that would otherwise get stuck in recursion.
  *  Log to the console exactly when it happens and then also try to get it into the log
@@ -177,6 +148,13 @@ extern void setSilentLogLevels();
  */
 extern void setSimpleLogLevels();
 
+#ifdef WIN32
+/**
+ * This sets a flag which tells the logger that alternate source labels should be used to indicate that the current process is a launcher.
+ */
+extern void setLauncherSource();
+#endif
+
 /**
  * Used for testing to set a pause into the next log entry made.
  *
@@ -184,9 +162,18 @@ extern void setSimpleLogLevels();
  */
 extern void setPauseTime(int pauseTime);
 
+/**
+ * Set to true to cause changes in internal buffer sizes to be logged.  Useful for debugging.
+ *
+ * @param log TRUE if changes should be logged.
+ */
+void setLogBufferGrowth(int log);
+
 /* * Logfile functions * */
 extern int isLogfileAccessed();
 
+extern int resolveDefaultLogFilePath();
+
 /**
  * Sets the log file to be used.  If the specified file is not absolute then
  *  it will be resolved into an absolute path.  If there are any problems with
@@ -194,17 +181,21 @@ extern int isLogfileAccessed();
  *  cause will be written to the existing log.
  *
  * @param log_file_path Log file to start using.
- * @param workingDir The current working directory, used for relative paths.
- *                   This will be NULL if this is part of the bootstrap process,
- *                   in which case we should not attempt to resolve the absolute
- *                   path.
- * @param preload TRUE if called as part of the preload process.  We use this to
- *                suppress double warnings.
+ * @param isConfigured  The value comes from the configuration file.
  *
  * @return TRUE if there were any problems.
  */
-extern int setLogfilePath( const TCHAR *log_file_path, const TCHAR *workingDir, int preload);
+extern int setLogfilePath( const TCHAR *log_file_path, int isConfigured);
 
+/**
+ * Returns the default logfile.
+ */
+extern const TCHAR *getDefaultLogfilePath();
+
+/**
+ * Returns a reference to the currect log file path.
+ *  This return value may be changed at any time if the log file is rolled.
+ */
 extern const TCHAR *getLogfilePath();
 
 /**
@@ -231,13 +222,21 @@ extern void setLogfileLevelInt(int log_file_level);
 extern int getLogfileLevelInt();
 extern void setLogfileLevel( const TCHAR *log_file_level );
 extern void setLogfileMaxFileSize( const TCHAR *max_file_size );
-extern void setLogfileMaxFileSizeInt(int max_file_size);
 extern void setLogfileMaxLogFiles(int max_log_files);
 extern void setLogfilePurgePattern(const TCHAR *pattern);
 extern void setLogfilePurgeSortMode(int sortMode);
 extern DWORD getLogfileActivity();
-extern void closeLogfile();
+
+/** Sets the auto flush log file flag. */
+extern void setLogfileAutoFlush(int autoFlush);
+
+/** Sets the auto close log file flag. */
 extern void setLogfileAutoClose(int autoClose);
+
+/** Closes the logfile if it is open. */
+extern void closeLogfile();
+
+/** Flushes any buffered logfile output to the disk. */
 extern void flushLogfile();
 
 /* * Console functions * */
@@ -246,6 +245,9 @@ extern void setConsoleLogLevelInt(int console_log_level);
 extern int getConsoleLogLevelInt();
 extern void setConsoleLogLevel( const TCHAR *console_log_level );
 extern void setConsoleFlush(int flush);
+#ifdef WIN32
+extern void setConsoleDirect(int direct);
+#endif
 extern void setConsoleFatalToStdErr(int toStdErr);
 extern void setConsoleErrorToStdErr(int toStdErr);
 extern void setConsoleWarnToStdErr(int toStdErr);
@@ -254,18 +256,29 @@ extern void setConsoleWarnToStdErr(int toStdErr);
 extern void setSyslogLevelInt(int loginfo_level);
 extern int getSyslogLevelInt();
 extern void setSyslogLevel( const TCHAR *loginfo_level );
+extern void setSyslogSplitMessages(int splitMessages);
+#ifdef WIN32
+extern void setSyslogRegister(int sysRegister);
+extern int getSyslogRegister();
+extern TCHAR* getSyslogEventSourceName();
+#endif
 #ifndef WIN32
 extern void setSyslogFacility( const TCHAR *loginfo_level );
 #endif
 extern void setSyslogEventSourceName( const TCHAR *event_source_name );
-extern int registerSyslogMessageFile();
-extern int unregisterSyslogMessageFile();
+extern void setThreadMessageBufferInitialSize(int initialValue);
+extern void disableSysLog(int silent);
+extern int syslogMessageFileRegistered(int silent);
+extern int registerSyslogMessageFile(int forceInstall, int silent);
+extern int unregisterSyslogMessageFile(int silent);
 
 
+extern void resetDuration();
+
 extern int getLowLogLevel();
 
 /* * General log functions * */
-extern int initLogging();
+extern int initLogging(void (*logFileChanged)(const TCHAR *logFile));
 extern int disposeLogging();
 extern void setUptime(int uptime, int flipped);
 extern void rollLogs();
@@ -274,6 +287,8 @@ extern int getLogLevelForName( const TCHAR *logLevelName );
 extern int getLogFacilityForName( const TCHAR *logFacilityName );
 #endif
 extern void logRegisterThread(int thread_id);
+extern void logRegisterFormatCallbacks(int (*countCallback)(const TCHAR format, size_t *reqSize), 
+                                       int (*printCallback)(const TCHAR format, size_t printSize, TCHAR** pBuffer));
 
 /**
  * The log_printf function logs a message to the configured log targets.
@@ -296,8 +311,6 @@ extern void log_printf( int source_id, int level, const TCHAR *lpszFmt, ... );
  */
 extern void log_printf_queue( int useQueue, int source_id, int level, const TCHAR *lpszFmt, ... );
 
-extern TCHAR* getLastErrorText();
-extern int getLastError();
 extern void maintainLogger();
 extern void invalidMultiByteSequence(const TCHAR *context, int id);
 #endif
