@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2017 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -56,6 +56,18 @@
 
 #include "property.h"
 
+#ifndef WIN32
+ /*
+  * Mac OSX 10.5 does not define the environ variable.  This is work around for that.
+  */
+ #ifdef MACOSX
+  #include <crt_externs.h>
+  #define environ *_NSGetEnviron();
+ #endif
+
+extern char** environ;
+#endif
+
 /* The following define will enable debug output of the code to parse the JVM output. */
 /*#define DEBUG_CHILD_OUTPUT*/
 
@@ -78,9 +90,13 @@
 #define WRAPPER_TIMER_FAST_THRESHOLD (2 * 24 * 3600 * 1000 / WRAPPER_TICK_MS) /* Default to 2 days. */
 #define WRAPPER_TIMER_SLOW_THRESHOLD (2 * 24 * 3600 * 1000 / WRAPPER_TICK_MS) /* Default to 2 days. */
 
-#define WRAPPER_BACKEND_TYPE_UNKNOWN 0 /* Unknown type. */
-#define WRAPPER_BACKEND_TYPE_SOCKET  1 /* Use a loopback socket to communicate. */
-#define WRAPPER_BACKEND_TYPE_PIPE    2 /* Use a pair of pipes to communicate. */
+#define WRAPPER_BACKEND_ERROR_NEXT     99 /* value returned when there is an error trying to connect to a socket and we should try the next way to connect */
+#define WRAPPER_BACKEND_TYPE_UNKNOWN   0 /* Unknown type. */
+#define WRAPPER_BACKEND_TYPE_SOCKET_V4 0x01
+#define WRAPPER_BACKEND_TYPE_SOCKET_V6 0x02
+#define WRAPPER_BACKEND_TYPE_SOCKET    (WRAPPER_BACKEND_TYPE_SOCKET_V4 | WRAPPER_BACKEND_TYPE_SOCKET_V6)
+#define WRAPPER_BACKEND_TYPE_PIPE      0x04 /* Use a pair of pipes to communicate. */
+#define WRAPPER_BACKEND_TYPE_AUTO      (WRAPPER_BACKEND_TYPE_SOCKET | WRAPPER_BACKEND_TYPE_PIPE)
 
 #define WRAPPER_WSTATE_STARTING  51 /* Wrapper is starting.  Remains in this state
                                      *  until the JVM enters the STARTED state or
@@ -138,6 +154,8 @@
                                      *  JVM will be killed and we will enter the STOPPED
                                      *  state. */
 #define WRAPPER_JSTATE_KILL      84 /* The Wrapper is about ready to kill the JVM process. */
+#define WRAPPER_JSTATE_KILLED    85 /* The Wrapper has requested termination of the JVM and 
+                                       confirmed the JVM is actually killed */
 
 /* Defined Action types.  Registered actions are negative.  Custom types are positive. */
 #define ACTION_LIST_END          0
@@ -155,10 +173,14 @@
 #define ACTION_ADVICE_NIL_SERVER -32
 #endif
 
+/* The following codes are passed through and referenced within the WrapperServiceActionEvent class.
+ *  They can be added, but not changed without the possibility of affecting user code. */
 #define WRAPPER_ACTION_SOURCE_CODE_FILTER                  1  /* Action originated with a filter. */
 #define WRAPPER_ACTION_SOURCE_CODE_COMMANDFILE             2  /* Action originated from a commandfile. */
 #define WRAPPER_ACTION_SOURCE_CODE_WINDOWS_SERVICE_MANAGER 3  /* Action originated from the Windows Service Manager. */
 #define WRAPPER_ACTION_SOURCE_CODE_ON_EXIT                 4  /* Action originated from an on_exit configuration. */
+#define WRAPPER_ACTION_SOURCE_CODE_SIGNAL                  5  /* Action originated from a signal. */
+#define WRAPPER_ACTION_SOURCE_CODE_PING_TIMEOUT            11 /* Action originated from a timeout. */
 
 /* Because of the way time is converted to ticks, the largest possible timeout that
  *  can be specified without causing 32-bit overflows is (2^31 / 1000) - 5 = 2147478
@@ -180,21 +202,30 @@ typedef unsigned int TICKS;
 typedef unsigned long TICKS;
 #endif
 
-#define WRAPPER_ENV_SOURCE_PARENT      1
-#define WRAPPER_ENV_SOURCE_WRAPPER     2
-#ifdef WIN32
-#define WRAPPER_ENV_SOURCE_REG_SYSTEM  4
-#define WRAPPER_ENV_SOURCE_REG_ACCOUNT 8
-#endif
-
 #ifdef WIN32
 /* Defines the maximum number of service manager control events that can be queued in a single loop. */
 #define CTRL_CODE_QUEUE_SIZE 26 /* Can enqueue one less than this count at any time. */
 #endif
 
+#define WRAPPER_JAVAIO_BUFFER_SIZE_SYSTEM_DEFAULT 0
+#define WRAPPER_JAVAIO_BUFFER_SIZE_MIN 1024
+#define WRAPPER_JAVAIO_BUFFER_SIZE_MAX (10 * 1024 * 1024)
+#define WRAPPER_JAVAIO_BUFFER_SIZE_DEFAULT (64 * 1024)
+
+/*#define DEBUG_PING_QUEUE*/
+#define WRAPPER_MAX_PENDING_PINGS 10
+typedef struct PendingPing PendingPing, *PPendingPing;
+struct PendingPing {
+    TICKS sentTicks;
+    TICKS timeoutTicks;
+    TICKS slowTicks;
+    PPendingPing nextPendingPing;
+};
+
 /* Type definitions */
 typedef struct WrapperConfig WrapperConfig;
 struct WrapperConfig {
+    TCHAR   *argBinary;             /* The name of the wrapper binary. */
     TCHAR   *argCommand;            /* The command used to launch the wrapper. */
     TCHAR   *argCommandArg;         /* The argument to the command used to launch the wrapper. */
     TCHAR   *argConfFile;           /* The name of the config file from the command line. */
@@ -211,6 +242,7 @@ struct WrapperConfig {
     int     backendType;            /* The type of the backend that the Wrapper and Java use to communicate. */
     int     configured;             /* TRUE if loadConfiguration has been called. */
     int     useSystemTime;          /* TRUE if the wrapper should use the system clock for timing, FALSE if a tick counter should be used. */
+    int     logBufferGrowth;        /* TRUE if changes to internal buffer sizes should be logged. */
     int     timerFastThreshold;     /* If the difference between the system time based tick count and the timer tick count ever falls by more than this value then a warning will be displayed. */
     int     timerSlowThreshold;     /* If the difference between the system time based tick count and the timer tick count ever grows by more than this value then a warning will be displayed. */
     int     useTickMutex;           /* TRUE if access to the tick count should be protected by a mutex. */
@@ -225,13 +257,17 @@ struct WrapperConfig {
     int     jvmPortMin;             /* Minimum port which the JVM should bind to when connecting back to the wrapper. */
     int     jvmPortMax;             /* Maximum port which the JVM should bind to when connecting back to the wrapper. */
     int     sock;                   /* Socket number. if open. */
+    TCHAR   *portAddress;
     TCHAR   *originalWorkingDir;    /* Original Wrapper working directory. */
     TCHAR   *workingDir;            /* Configured working directory. */
     TCHAR   *configFile;            /* Name of the configuration file */
     int     commandLogLevel;        /* The log level to use when logging the java command. */
+    int     printJVMVersion;        /* tells the Wrapper to create a temp JVM to query the version, before starting the java application */
 #ifdef WIN32
+    TCHAR   *jvmVersionCommand;     /* Command used to launch the JVM and request its version */
     TCHAR   *jvmCommand;            /* Command used to launch the JVM */
 #else /* UNIX */
+    TCHAR   **jvmVersionCommand;    /* Command used to launch the JVM and request its version */
     TCHAR   **jvmCommand;           /* Command used to launch the JVM */
 #endif
     int     detachStarted;          /* TRUE if the JVM process should be detached once it has started. */
@@ -244,11 +280,20 @@ struct WrapperConfig {
     int     cpuTimeout;             /* Number of seconds without CPU before the JVM will issue a warning and extend timeouts */
     int     startupTimeout;         /* Number of seconds the wrapper will wait for a JVM to startup */
     int     pingTimeout;            /* Number of seconds the wrapper will wait for a JVM to reply to a ping */
+    int     pingAlertThreshold;     /* Number of seconds without a ping response that the Wrapper will start to warn about a slow ping. */
+    int     pingAlertLogLevel;      /* Long level at which slow ping notices are logged. */
     int     pingInterval;           /* Number of seconds between pinging the JVM */
     int     pingIntervalLogged;     /* Number of seconds between pings which can be logged to debug output. */
+    int     *pingActionList;        /* The action list to take when a ping timeout is detected. */
+    int     pingTimedOut;
     int     shutdownTimeout;        /* Number of seconds the wrapper will wait for a JVM to shutdown */
     int     jvmExitTimeout;         /* Number of seconds the wrapper will wait for a JVM to process to terminate */
     int     jvmCleanupTimeout;      /* Number of seconds the wrapper will allow for its post JVM shudown cleanup. */
+    int     jvmTerminateTimeout;    /* Number of seconds the wrapper will allow for the JVM to respond to TerminateProcess request. */
+    int     jvmSilentKill;          /* TRUE if the JVM should be silently killed at the next opportunity. */
+#ifdef WIN32
+    int     javaIOBufferSize;       /* Size of the pipe buffer to use for java I/O. */
+#endif
     int     useJavaIOThread;        /* If TRUE then a dedicated thread will be used to process console output form the JVM. */
     int     pauseThreadMain;        /* Number of seconds to pause the main thread on its next loop.  Only used for testing. */
     int     pauseThreadTimer;       /* Number of seconds to pause the timer thread on its next loop.  Only used for testing. */
@@ -291,14 +336,25 @@ struct WrapperConfig {
     int     isPageFaultOutputEnabled;/* TRUE if detailed PageFault output should be included in status output. */
     int     pageFaultOutputInterval;/* Interval in seconds at which PageFault usage is logged. */
     TICKS   pageFaultOutputTimeoutTicks; /* Tick count at which PageFault will next be logged. */
-    int     logfileInactivityTimeout; /* The number of seconds of inactivity before the logfile will be closed. */
-    TICKS   logfileInactivityTimeoutTicks; /* Tick count at which the logfile will be considered inactive and closed. */
-    int     isShutdownHookDisabled; /* TRUE if set in the configuration file */
+    int     logfileFlushTimeout;    /* The number of seconds before the logfile will be flushed. */
+    TICKS   logfileFlushTimeoutTicks; /* Tick count at which the logfile will be flushed. */
+    int     logfileFlushTimeoutTicksSet; /* TRUE if logfileFlushTimeoutTicks is set. */
+    int     logfileCloseTimeout;    /* The number of seconds of inactivity before the logfile will be closed. */
+    TICKS   logfileCloseTimeoutTicks; /* Tick count at which the logfile will be considered inactive and closed. */
+    int     logfileCloseTimeoutTicksSet; /* TRUE if logfileCloseTimeoutTicks is set. */
+    int     isTestsDisabled;        /* TRUE if the use of tests in the WrapperManager class should be disabled. */
+    int     isShutdownHookDisabled; /* TRUE if the use of a shutdown hook by the WrapperManager class should be disabled. */
+    int     isForcedShutdownDisabled; /* TRUE if forced shutdowns are disabled. */
     int     startupDelayConsole;    /* Delay in seconds before starting the first JVM in console mode. */
     int     startupDelayService;    /* Delay in seconds before starting the first JVM in service mode. */
     int     exitCode;               /* Code which the wrapper will exit with */
+    int     errorExitCode;          /* Code which the wrapper will exit with when there is an error. */
     int     exitRequested;          /* TRUE if the current JVM should be shutdown. */
     int     restartRequested;       /* WRAPPER_RESTART_REQUESTED_NO, WRAPPER_RESTART_REQUESTED_AUTOMATIC, or WRAPPER_RESTART_REQUESTED_CONFIGURED if the another JVM should be launched after the current JVM is shutdown. Only set if exitRequested is set. */
+    int     shutdownActionTriggered; /* TRUE if any action causing a JVM exit was triggered by the Wrapper (see the WRAPPER_ACTION_SOURCE_CODE_* definitions above).
+                                     *  This flag indicates that the action specified by wrapper.on_exit should be ignored. It must be reset after the JVM is
+                                     *  restarted so that wrapper.on_exit can continue to catch JVM exit codes. */
+    TCHAR*  shutdownActionPropertyName; /* Keep the name of the property that originated the shudown for logging. */
     int     stoppedPacketReceived;  /* TRUE if the STOPPED packet was received before a restart. */
     int     restartPacketReceived;  /* TRUE if the RESTART packet was received before a restart. */
     int     jvmRestarts;            /* Number of times that a JVM has been launched since the wrapper was started. */
@@ -319,6 +375,7 @@ struct WrapperConfig {
     int     *outputFilterAllowWildFlags; /* Array of output filter flags that say whether or not wild cards in the filter can be processed. */
     size_t  *outputFilterMinLens;   /* Array of the minimum text lengths that could possibly match the specified filter.  Only used if it contains wildcards. */
     TCHAR   *pidFilename;           /* Name of file to store wrapper pid in */
+    int     pidFileStrict;          /* TRUE if a preexisting pid file should cause an error. */
     TCHAR   *lockFilename;          /* Name of file to store wrapper lock in */
     TCHAR   *javaPidFilename;       /* Name of file to store jvm pid in */
     TCHAR   *javaIdFilename;        /* Name of file to store jvm id in */
@@ -367,13 +424,20 @@ struct WrapperConfig {
     int     ntServiceInteractive;   /* Should the service be allowed to interact with the desktop? */
     int     ntHideJVMConsole;       /* Should the JVMs Console window be hidden when run as a service.  True by default but GUIs will not be visible for JVMs prior to 1.4.0. */
     int     ntHideWrapperConsole;   /* Should the Wrapper Console window be hidden when run as a service. */
+    HINSTANCE wrapperHInstance;     /* The HINSTANCE of the Wrapper process. */
     int     wrapperConsoleHide;     /* True if the Wrapper Console window should be hidden. */
-    HWND    wrapperConsoleHandle;   /* Pointer to the Wrapper Console handle if it exists.  This will only be set if the console was allocated then hidden. */
+    HWND    wrapperConsoleHWND;     /* The HWND of the Wrapper's console if it was located. */
     int     wrapperConsoleVisible;  /* True if the Wrapper Console window is visible. */
     HWND    jvmConsoleHandle;       /* Pointer to the JVM Console handle if it exists. */
     int     jvmConsoleVisible;      /* True if the JVM Console window is visible. */
     int     ntAllocConsole;         /* True if a console should be allocated for the Service. */
     int     generateConsole;        /* Make sure that a console is always generated to support thread dumps */
+#ifndef WINIA
+    int     ntPreshutdown;          /* True if the service accepts SERVICE_CONTROL_PRESHUTDOWN controle code, False if it accepts SERVICE_CONTROL_SHUTDOWN. */
+    int     ntPreshutdownTimeout;   /* Number of seconds that the service controler will wait after sending the SERVICE_CONTROL_PRESHUTDOWN notification. */
+#endif
+    int     ntShutdownWaitHint;     /* Delay in seconds requested by the Wrapper when it reports a STOPPING or PAUSING status to the service controler. */
+    int     ntStartupWaitHint;      /* Delay in seconds requested by the Wrapper when it reports a STARTING or RESUMING status to the service controler. */
     int     threadDumpControlCode;  /* Control code which can be used to trigger a thread dump. */
 #else /* UNIX */
     int     daemonize;              /* TRUE if the process  should be spawned as a daemon process on launch. */
@@ -382,6 +446,13 @@ struct WrapperConfig {
     int     signalUSR2Mode;         /* Controls what happens when the Wrapper receives a USR2 signal. */
     int     jvmStopped;             /* Flag which remembers the the stopped state of the JVM process. */
 #endif
+    
+    int pendingPingQueueOverflow;   /* Flag which is set to true if the PendingPingQueue overflows the limit of WRAPPER_MAX_PENDING_PINGS. */
+    int pendingPingQueueOverflowEmptied; /* Flag which is set when the queue size is reduced to 0 after having overflowed. */
+    int pendingPingCount;           /* Number of PendingPing events in the list. */
+    PPendingPing firstPendingPing;  /* Pointer to the first PendingPing in the list. */
+    PPendingPing firstUnwarnedPendingPing; /* Pointer to the first PendingPing in the list for which a slow warning has not been logged. */
+    PPendingPing lastPendingPing;   /* Pointer to the last PendingPing in the list. */
 
 #ifdef WIN32
     int     ctrlEventCTRLCTrapped;  /* CTRL_C_EVENT trapped. */
@@ -396,6 +467,9 @@ struct WrapperConfig {
     int     ctrlCodeContinueTrapped;/* SERVICE_CONTROL_CONTINUE was trapped. */
     int     ctrlCodeStopTrapped;    /* SERVICE_CONTROL_STOP was trapped. */
     int     ctrlCodeShutdownTrapped;/* SERVICE_CONTROL_SHUTDOWN was trapped. */
+ #ifndef WINIA
+    int     ctrlCodePreShutdownTrapped;/* SERVICE_CONTROL_PRESHUTDOWN was trapped. */
+ #endif
     int     ctrlCodeDumpTrapped;    /* The configured thread dump control code was trapped. */
 #else
     int     signalInterruptTrapped; /* SIGINT was trapped. */
@@ -408,10 +482,14 @@ struct WrapperConfig {
 #endif
 };
 
+
 #define WRAPPER_SIGNAL_MODE_IGNORE   (char)100
 #define WRAPPER_SIGNAL_MODE_RESTART  (char)101
 #define WRAPPER_SIGNAL_MODE_SHUTDOWN (char)102
 #define WRAPPER_SIGNAL_MODE_FORWARD  (char)103
+#define WRAPPER_SIGNAL_MODE_PAUSE    (char)104
+#define WRAPPER_SIGNAL_MODE_RESUME   (char)105
+#define WRAPPER_SIGNAL_MODE_CLOSE_LOGFILE (char)106
 
 #define WRAPPER_MSG_START         (char)100
 #define WRAPPER_MSG_STOP          (char)101
@@ -424,19 +502,23 @@ struct WrapperConfig {
 #define WRAPPER_MSG_KEY           (char)110
 #define WRAPPER_MSG_BADKEY        (char)111
 #define WRAPPER_MSG_LOW_LOG_LEVEL (char)112
-#define WRAPPER_MSG_PING_TIMEOUT  (char)113 /* No longer used. */
+#define WRAPPER_MSG_PING_TIMEOUT  (char)113 /* No longer used.  But keep reserved to avoid future problems. */
 #define WRAPPER_MSG_SERVICE_CONTROL_CODE (char)114
 #define WRAPPER_MSG_PROPERTIES    (char)115
 /** Log commands are actually 116 + the LOG LEVEL (LEVEL_UNKNOWN ~ LEVEL_NONE), (116 ~ 124). */
 #define WRAPPER_MSG_LOG           (char)116
 #define WRAPPER_MSG_LOGFILE       (char)134
-#define WRAPPER_MSG_APPEAR_ORPHAN (char)137 /* No longer used. */
+#define WRAPPER_MSG_APPEAR_ORPHAN (char)137 /* No longer used.  But keep reserved to avoid future problems. */
 #define WRAPPER_MSG_PAUSE         (char)138
 #define WRAPPER_MSG_RESUME        (char)139
 #define WRAPPER_MSG_GC            (char)140
+
 #define WRAPPER_PROCESS_DOWN      200
 #define WRAPPER_PROCESS_UP        201
 
+/* default timeout (in second) when executing child process to print java version */
+#define DEFAULT_JAVA_VERSION_TIMEOUT    10
+
 extern WrapperConfig *wrapperData;
 extern Properties    *properties;
 
@@ -483,6 +565,24 @@ extern int wrapperProtocolRead();
 /******************************************************************************
  * Utility Functions
  *****************************************************************************/
+
+/**
+ * Does any necessary post processing on the command string.
+ *  This function assumes that command has been malloced.  It will either return
+ *  the string as is, or return a modified string.  When a modified string is
+ *  returned the orignal command buffer will always be freed.
+ *
+ * 1) Replace the first instance of the %WRAPPER_COMMAND_FILLER_N% environment
+ *    variable so that the total command length will be equal to or greater than
+ *    the length specified by N.  The padding will be the length plus a series of
+ *    Xs terminated by a single Y.  This is mainly for testing.
+ *
+ * @param command The original command.
+ *
+ * @return The modifed command.
+ */
+extern TCHAR *wrapperPostProcessCommandElement(TCHAR *command);
+
 /**
  * Test function to pause the current thread for the specified amount of time.
  *  This is used to test how the rest of the Wrapper behaves when a particular
@@ -518,12 +618,18 @@ extern int wrapperWildcardMatch(const TCHAR *text, const TCHAR *pattern, size_t
  */
 extern size_t wrapperGetMinimumTextLengthForPattern(const TCHAR *pattern);
 
+/**
+ * Trims any whitespace from the beginning and end of the in string
+ *  and places the results in the out buffer.  Assumes that the out
+ *  buffer is at least as large as the in buffer. */
+extern void trim(const TCHAR *in, TCHAR *out);
+
 /**
  * Returns a constant text representation of the specified Wrapper State.
  *
  * @param wState The Wrapper State whose name is being requested.
  *
- * @return Thre requested Wrapper State.
+ * @return The requested Wrapper State.
  */
 extern const TCHAR *wrapperGetWState(int wState);
 
@@ -532,7 +638,7 @@ extern const TCHAR *wrapperGetWState(int wState);
  *
  * @param jState The Java State whose name is being requested.
  *
- * @return Thre requested Java State.
+ * @return The requested Java State.
  */
 extern const TCHAR *wrapperGetJState(int jState);
 
@@ -544,6 +650,7 @@ extern int initializeStartup();
 extern void disposeStartup();
 #endif
 extern void disposeJavaIO();
+extern int initializeTimer();
 extern void disposeTimer();
 
 extern int showHostIds(int logLevel);
@@ -566,26 +673,26 @@ extern int *wrapperGetActionListForNames(const TCHAR *actionNameList, const TCHA
  *                   Negative values are standard actions, positive are user
  *                   custom events.
  * @param triggerMsg The reason the actions are being fired.
- * @param actionCode Tracks where the action originated.
+ * @param actionSourceCode Tracks where the action originated.
+ * @param actionPropertyIndex Index of the property where the action was configured. Ignored if the type of action is not configured with a <n>-component property.
  * @param logForActionNone Flag stating whether or not a message should be logged
  *                         for the NONE action.
  * @param exitCode Error code to use in case the action results in a shutdown.
  */
-extern void wrapperProcessActionList(int *actionList, const TCHAR *triggerMsg, int actionCode, int logForActionNone, int exitCode);
+extern void wrapperProcessActionList(int *actionList, const TCHAR *triggerMsg, int actionSourceCode, int actionPropertyIndex, int logForActionNone, int exitCode);
 
 extern void wrapperAddDefaultProperties();
 
-extern int wrapperLoadConfigurationProperties();
+extern int wrapperLoadConfigurationProperties(int preload);
 
 extern void wrapperGetCurrentTime(struct timeb *timeBuffer);
 
+extern void updateStringValue(TCHAR **ptr, const TCHAR *value);
+
 #ifdef WIN32
 
 extern void wrapperInitializeProfileCounters();
 extern void wrapperDumpPageFaultUsage();
-
-extern void updateStringValue(TCHAR **ptr, const TCHAR *value);
-
 extern TCHAR** wrapperGetSystemPath();
 extern int wrapperGetJavaHomeFromWindowsRegistry(TCHAR *javaHome);
 #endif
@@ -638,14 +745,20 @@ extern void wrapperJVMProcessExited(TICKS nowTicks, int exitCode);
 
 /**
  * Read and process any output from the child JVM Process.
- * Most output should be logged to the wrapper log file.
  *
- * This function will only be allowed to run for 250ms before returning.  This is to
- *  make sure that the main loop gets CPU.  If there is more data in the pipe, then
- *  the function returns TRUE, otherwise FALSE.  This is a hint to the mail loop not to
- *  sleep.
+ * When maxTimeMS is non-zero this function will only be allowed to run for that maximum
+ *  amount of time.  This is done to make sure the calling function is allowed CPU for
+ *  other activities.   When timing out for this reason when there is more data in the
+ *  pipe, this function will return TRUE to let the calling code know that it should
+ *  not to any unnecessary sleeps.  Otherwise FALSE will be returned.
+ *
+ * @param maxTimeMS The maximum number of milliseconds that this function will be allowed
+ *                  to run without returning.  In reality no new reads will happen after
+ *                  this time, but actual processing may take longer.
+ *
+ * @return TRUE if the calling code should call this function again as soon as possible.
  */
-extern int wrapperReadChildOutput();
+extern int wrapperReadChildOutput(int maxTimeMS);
 
 /**
  * Changes the current Wrapper state.
@@ -692,17 +805,18 @@ BOOL extern duplicateSTD();
 BOOL extern myShellExec(HWND hwnd, LPCTSTR pszVerb, LPCTSTR pszPath, LPCTSTR pszParameters, LPCTSTR pszDirectory, TCHAR* namedPipeName);
 BOOL extern runElevated( __in LPCTSTR pszPath, __in_opt LPCTSTR pszParameters, __in_opt LPCTSTR pszDirectory, TCHAR* namedPipeName);
 BOOL extern isElevated();
+BOOL extern isWin10OrHigher();
 BOOL extern isVista();
+BOOL extern isWinXP();
 extern void wrapperMaintainControlCodes();
+int wrapperServiceStatus(int consoleOutput);
+#define OSBUFSIZE 256
+BOOL GetOSDisplayString(TCHAR** pszOS);
 #else
 extern void wrapperMaintainSignals();
+extern TCHAR* findPathOf(const TCHAR *exe, const TCHAR* name);
 #endif
 
-/**
- * Gets the error code for the last operation that failed.
- */
-extern int wrapperGetLastError();
-
 /**
  * Execute initialization code to get the wrapper set up.
  */
@@ -753,8 +867,10 @@ extern void wrapperPauseBeforeExecute();
 
 /**
  * Launches a JVM process and store it internally
+ *
+ * @return TRUE if there were any problems.  When this happens the Wrapper will not try to restart.
  */
-extern void wrapperExecute();
+extern int wrapperExecute();
 
 /**
  * Returns a tick count that can be used in combination with the
@@ -791,15 +907,20 @@ extern void wrapperDumpCPUUsage();
  * Immediately kill the JVM process and set the JVM state to
  *  WRAPPER_JSTATE_DOWN.
  */
-extern void wrapperKillProcessNow();
+extern int wrapperKillProcessNow();
+
 
 /**
  * Puts the Wrapper into a state where the JVM will be killed at the soonest
  *  possible opportunity.  It is necessary to wait a moment if a final thread
- *  dump is to be requested.  This call wll always set the JVM state to
+ *  dump is to be requested.  This call will always set the JVM state to
  *  WRAPPER_JSTATE_KILLING.
+ *
+ * @param silent TRUE to skip messages saying that the JVM did not exit on request.
+ *               This is useful in certain cases where we kill the JVM without trying
+ *               to shut it down cleanly.
  */
-extern void wrapperKillProcess();
+extern void wrapperKillProcess(int silent);
 
 /**
  * Launch the wrapper as a console application.
@@ -814,16 +935,16 @@ extern int wrapperRunService();
 /**
  * Used to ask the state engine to pause the JVM and Wrapper
  *
- * @param actionCode Tracks where the action originated.
+ * @param actionSourceCode Tracks where the action originated.
  */
-extern void wrapperPauseProcess(int actionCode);
+extern void wrapperPauseProcess(int actionSourceCode);
 
 /**
  * Used to ask the state engine to resume the JVM and Wrapper
  *
- * @param actionCode Tracks where the action originated.
+ * @param actionSourceCode Tracks where the action originated.
  */
-extern void wrapperResumeProcess(int actionCode);
+extern void wrapperResumeProcess(int actionSourceCode);
 
 /**
  * Detaches the Java process so the Wrapper will if effect forget about it.
@@ -841,6 +962,13 @@ extern void wrapperDetachJava();
  */
 extern void wrapperStopProcess(int exitCode, int force);
 
+/**
+ * Depending on the current state, we want to change the exact message displayed when restarting the JVM.
+ *
+ * The logic here needs to match that in wrapperRestartProcess.
+ */
+extern const TCHAR *wrapperGetRestartProcessMessage();
+
 /**
  * Used to ask the state engine to shut down the JVM.
  */
@@ -848,8 +976,10 @@ extern void wrapperRestartProcess();
 
 /**
  * Sends a command off to the JVM asking it to perform a garbage collection sweep.
+ *
+ * @param actionSourceCode Tracks where the action originated.
  */
-extern void wrapperRequestJVMGC();
+extern void wrapperRequestJVMGC(int actionSourceCode);
 
 /**
  * Loops over and strips all double quotes from prop and places the
@@ -942,21 +1072,63 @@ extern TICKS wrapperAddToTicks(TICKS start, int seconds);
  * Sets the working directory of the Wrapper to the specified directory.
  *  The directory can be relative or absolute.
  * If there are any problems then a non-zero value will be returned.
+ *
+ * @param dir Directory to change to.
+ * @param logErrors TRUE if errors should be logged.
  */
-extern int wrapperSetWorkingDir(const TCHAR* dir);
+extern int wrapperSetWorkingDir(const TCHAR* dir, int logErrors);
 
 /******************************************************************************
  * Protocol callback functions
  *****************************************************************************/
 extern void wrapperLogSignaled(int logLevel, TCHAR *msg);
 extern void wrapperKeyRegistered(TCHAR *key);
-extern void wrapperPingResponded();
+
+/**
+ * Called when a ping is first determined to be slower than the wrapper.ping.alert.threshold.
+ *  This will happen before it has actually been responded to.
+ */
+extern void wrapperPingSlow();
+
+/**
+ * Called when a ping is responded to, but was slower than the wrapper.ping.alert.threshold.
+ *
+ * @param tickAge The number of seconds it took to respond.
+ */
+extern void wrapperPingRespondedSlow(int tickAge);
+
+/**
+ * Called when a ping response is received.
+ *
+ * @param pingSendTicks Time in ticks when the ping was originally sent.
+ * @param queueWarnings TRUE if warnings about the queue should be logged, FALSE if the ping response did not contain a time.
+ */
+extern void wrapperPingResponded(TICKS pingSendTicks, int queueWarnings);
+
+extern void wrapperPingTimeoutResponded();
 extern void wrapperStopRequested(int exitCode);
 extern void wrapperRestartRequested();
 extern void wrapperStopPendingSignaled(int waitHint);
 extern void wrapperStoppedSignaled();
 extern void wrapperStartPendingSignaled(int waitHint);
 extern void wrapperStartedSignaled();
-extern int file_exists(const TCHAR * filename);
 
+/******************************************************************************
+ * Inner types and methods for loading Wrapper configuration.
+ *****************************************************************************/
+
+/* Callback parameter for loading the file specified by
+   wrapper.java.additional_file configuration property. */
+typedef struct LoadParameterFileCallbackParam LoadParameterFileCallbackParam;
+struct LoadParameterFileCallbackParam {
+    int stripQuote;  /* Value of wrapper.java.additional_file.stripquotes property */
+    TCHAR **strings; /* Array of character strings to which configurations are loaded */
+    int index;       /* Index of a string in `strings' to which a configuration
+                        are copied next */
+    int isJVMParam;  /* Indicates whether the loaded parameters will be JVM parameters */
+};
+
+#ifdef CUNIT
+extern void tsJAP_testJavaAdditionalParamSuite(void);
+#endif /* CUNIT */
 #endif
