@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2013 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2017 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -13,42 +13,47 @@
  #define _LOCALIZE
  #include <stdio.h>
 
-
  #ifndef WIN32
+    
+   #define MBSTOWCS_QUERY_LENGTH 0
 
   #ifdef UNICODE
    #include <wchar.h>
-  #ifdef _sntprintf
-   #undef _sntprintf
-  #endif
+   #ifdef _sntprintf
+    #undef _sntprintf
+   #endif
 
-  #include <stdarg.h>
-  #include <stdlib.h>
-  #include <unistd.h>
-  #include <sys/types.h>
-  #include <sys/stat.h>
-  #include <locale.h>
-  #include <syslog.h>
-  #include <time.h>
-  #include <wctype.h>
+   #include <stdarg.h>
+   #include <stdlib.h>
+   #include <unistd.h>
+   #include <sys/types.h>
+   #include <sys/stat.h>
+   #include <locale.h>
+   #include <syslog.h>
+   #include <time.h>
+   #include <wctype.h>
   
-  #define __max(x,y) (((x) > (y)) ? (x) : (y))
-  #define __min(x,y) (((x) < (y)) ? (x) : (y))
+   #define __max(x,y) (((x) > (y)) ? (x) : (y))
+   #define __min(x,y) (((x) < (y)) ? (x) : (y))
 
-  #if defined(SOLARIS) || defined(HPUX)
-   #define WRAPPER_USE_PUTENV
-  #endif
+   #if defined(SOLARIS) || defined(HPUX)
+    #define WRAPPER_USE_PUTENV
+   #endif
 
-
-  #if defined(MACOSX) || defined(HPUX) || defined(FREEBSD) || defined(SOLARIS)
-   #ifndef wcscasecmp
-extern int wcscasecmp(const wchar_t* s1, const wchar_t* s2);
-    #define ECSCASECMP
+   #ifdef FREEBSD
+    #include <dlfcn.h>
    #endif
-  #endif
 
 
-  #define TEXT(x) L##x
+   #if defined(MACOSX) || defined(HPUX) || defined(FREEBSD) || defined(SOLARIS)
+    #ifndef wcscasecmp
+extern int wcscasecmp(const wchar_t* s1, const wchar_t* s2);
+     #define ECSCASECMP
+    #endif
+   #endif
+
+
+   #define TEXT(x) L##x
 typedef wchar_t TCHAR;
 typedef wchar_t _TUCHAR;
 
@@ -93,7 +98,6 @@ extern TCHAR* _tgetcwd(TCHAR *buf, size_t size);
 #define _tmkdir       _wmkdir
 #define _tmktemp      _wmktemp
 extern int _topen(const TCHAR *path, int oflag, mode_t mode);
-#define _tpopen       _wpopen
 #define _puttch       _putwch
 #if defined(WRAPPER_USE_PUTENV)
 extern int _tputenv(const TCHAR *string);
@@ -122,9 +126,12 @@ extern int _tstat(const wchar_t* filename, struct stat *buf);
 #define _tstati64     _wstati64
 #define _tstrdate     _wstrdate
 #define _tcsdec       _wcsdec
-#define _tcsdup       _wcsdup
+#define _tcsdup       wcsdup   /* replaced _wcsdup by wcsdup - but both not supported on Zos */
 #define _tcsicmp      wcscasecmp
-extern wchar_t* _trealpath(const wchar_t* file_name, wchar_t *resolved_name) ;
+/* Intentionally do not allow use of _trealpath because it does not specify a buffer length.
+ * #define _trealpath
+ * Define our own _trealpathN below. */
+extern wchar_t* _trealpathN(const wchar_t* fileName, wchar_t *resolvedName, size_t resolvedNameSize);
 #define _tcsicoll     _wcsicoll
 #define _tcsinc       _wcsinc
 #define _tcslwr       _wcslwr
@@ -169,13 +176,21 @@ extern int _vsntprintf(wchar_t *ws, size_t n, const wchar_t *format, va_list arg
 #define _tctime       _wctime
 #define _fgettc       fgetwc
 #define _fgetts       fgetws
-extern FILE * _tfopen(const wchar_t* file, const wchar_t* mode) ;
+extern FILE* _tfopen(const wchar_t* file, const wchar_t* mode);
+extern FILE* _tpopen(const wchar_t* command, const wchar_t* mode);
 #define _fputtc       fputwc
 #define _fputts       fputws
 #define _tfreopen     _wfreopen
 #define _ftscanf      fwscanf
 #define _gettc        getwc
 #define _gettchar     getwchar
+/**
+ * This Wrapper function internally does a malloc to generate the
+ *  Wide-char version of the return string.  This must be freed by the caller.
+ *  Only needed inside the following:
+ *  #if !defined(WIN32) && defined(UNICODE)
+ *  #endif
+ */
 extern TCHAR * _tgetenv ( const TCHAR * name );
 #define _getts        getws
 #define _istalnum     iswalnum
@@ -289,7 +304,6 @@ typedef unsigned char _TUCHAR;
 #define _tmkdir       _mkdir
 #define _tmktemp      _mktemp
 #define _topen        open
-#define _tpopen       _popen
 #define _puttch       _putch
 /*#define _tputenv      putenv*/
 #define _tsetenv      setenv
@@ -314,11 +328,14 @@ typedef unsigned char _TUCHAR;
 #define _tstati64     _stati64
 #define _tstrdate     _strdate
 #define _tcsdec       _strdec
-#define _tcsdup       _strdup
+#define _tcsdup       strdup    /* replaced _strdup by strdup */
 #define _tcsicmp      strcasecmp
 #define _tcsicoll     _stricoll
 #define _tcsinc       _strinc
-#define _trealpath    realpath
+/* Intentionally do not allow use of _trealpath because it does not specify a buffer length.
+ * #define _trealpath    realpath
+ * Define our own _trealpathN below. */
+#define _trealpathN(fileName, resolvedName, resolvedNameSize)    realpath(fileName, resolvedName)
 #define _tcslwr       _strlwr
 #define _tcsnbcnt     _strncnt
 #define _tcsnccnt     _strncnt
@@ -358,6 +375,7 @@ typedef unsigned char _TUCHAR;
 #define _fgettc       fgetc
 #define _fgetts       fgets
 #define _tfopen       fopen
+#define _tpopen       popen
 #define _ftprintf     fprintf
 #define _fputtc       fputc
 #define _fputts       fputs
@@ -436,4 +454,99 @@ typedef unsigned char _TUCHAR;
 #include <sys/stat.h>
 extern int multiByteToWideChar(const char *multiByteChars, int encoding, TCHAR **outputBufferW, int localizeErrorMessage);
 #endif
+/* Define boolean constants. */
+#ifndef TRUE
+ #define TRUE -1
 #endif
+
+#ifndef FALSE
+ #define FALSE 0
+#endif
+
+extern TCHAR* toLower(TCHAR* value);
+
+#ifdef _LIBICONV_VERSION
+ #ifdef AIX /* the AIX version of iconv.h doesn't have _LIBICONV_VERSION (need to check the other platforms) */
+  #define USE_LIBICONV_GNU
+ #endif
+#endif
+
+#ifdef FREEBSD
+/**
+ * Get the name of the iconv library that was loaded.
+ *
+ * @return the name of the iconv library (wide chars).
+ */
+extern TCHAR* getIconvLibName();
+
+/**
+ * Tries to load libiconv and then fallback in FreeBSD.
+ * Unfortunately we can not do any pretty logging here as iconv is
+ *  required for all of that to work.
+ *
+ * @return TRUE if there were any problems, FALSE otherwise.
+ */
+extern int loadIconvLibrary();
+#endif
+
+/**
+ * Define a cross platform way to compare strings while ignoring case.
+ */
+#ifdef WIN32
+#define strIgnoreCaseCmp _stricmp
+#else
+#define strIgnoreCaseCmp strcasecmp
+#endif
+#define strcmpIgnoreCase(str1, str2) _tcsicmp(str1, str2)
+
+/**
+ * Function to get the system encoding name/number for the encoding
+ * of the conf file
+ *
+ * @para String holding the encoding from the conf file
+ *
+ * @return TRUE if not found, FALSE otherwise
+ *
+ */
+#ifdef WIN32
+extern int getEncodingByName(char* encodingMB, int *encoding);
+#else
+extern int getEncodingByName(char* encodingMB, char** encoding);
+#endif
+
+/**
+ * Gets the error code for the last operation that failed.
+ */
+extern int wrapperGetLastError();
+
+/*
+ * Corrects a path in place by replacing all '/' characters with '\'
+ *  on Windows platforms.  Does nothing on NIX platforms.
+ *
+ * filename - Filename to be modified.  Could be null.
+ */
+extern void wrapperCorrectWindowsPath(TCHAR *filename);
+#endif
+
+
+/* Helper defines used to help trace where certain calls are being made. */
+/*#define DEBUG_MBSTOWCS*/
+#ifdef DEBUG_MBSTOWCS
+ #ifdef WIN32
+  #define mbstowcs(x,y,z) mbstowcs(x,y,z); wprintf(L"%S:%d:%S mbstowcs(%S, %S, %S) -> mbstowcs(%p, \"%S\", %d)\n", __FILE__, __LINE__, __FUNCTION__, #x, #y, #z, (void *)x, y, (int)z)
+ #else
+  #define mbstowcs(x,y,z) mbstowcs(x,y,z); wprintf(L"%s:%d:%s mbstowcs(%s, %s, %s) -> mbstowcs(%p, \"%s\", %d)\n", __FILE__, __LINE__, __FUNCTION__, #x, #y, #z, (void *)x, y, (int)z)
+ #endif
+#endif
+
+/*#define DEBUG_MALLOC*/
+#ifdef DEBUG_MALLOC
+ extern void *malloc2(size_t size, const char *file, int line, const char *func, const char *sizeVar);
+ #define malloc(x) malloc2(x, __FILE__, __LINE__, __FUNCTION__, #x)
+ #ifdef WIN32
+  #define free(x) wprintf(L"%S:%d:%S free(%S) -> free(%p)\n", __FILE__, __LINE__, __FUNCTION__, #x, (void *)x); free(x)
+ #else
+  #define free(x) wprintf(L"%s:%d:%s free(%s) -> free(%p)\n", __FILE__, __LINE__, __FUNCTION__, #x, (void *)x); free(x)
+ #endif
+#endif
+
