@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2017 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -35,6 +35,10 @@
 
 #ifndef WIN32
 
+#ifdef LINUX
+ #include <features.h>
+#endif
+#include <limits.h>
 #include <wchar.h>
 #include <netdb.h>
 #include <stdio.h>
@@ -57,16 +61,17 @@
 #include "wrapperinfo.h"
 #include "property.h"
 #include "logger.h"
+#include "wrapper_file.h"
 
 #include <sys/resource.h>
 #include <sys/time.h>
 
 #if defined(IRIX)
-#define PATH_MAX FILENAME_MAX
+ #define PATH_MAX FILENAME_MAX
 #endif
 
 #ifndef USE_USLEEP
-#include <time.h>
+ #include <time.h>
 #endif
 
 #ifndef getsid
@@ -84,6 +89,12 @@ int pipedes[2] = {-1, -1};
 #define PIPE_READ_END 0
 #define PIPE_WRITE_END 1
 
+/**
+ * maximum length for a user name should be 8, 
+ * but according to 'man useradd' it may be 32
+ */
+#define MAX_USER_NAME_LENGTH 32
+
 TCHAR wrapperClasspathSeparator = TEXT(':');
 
 int javaIOThreadSet = FALSE;
@@ -154,23 +165,25 @@ void appExit(int exitCode, int argc, TCHAR** argv) {
     exit(exitCode);
 }
 
-/**
- * Gets the error code for the last operation that failed.
- */
-int wrapperGetLastError() {
-    return errno;
-}
-
 /**
  * Writes a PID to disk.
  *
  * filename: File to write to.
  * pid: pid to write in the file.
+ * strict: If true then an error will be reported and the call will fail if the
+ *         file already exists.
+ *
+ * return 1 if there was an error, 0 if Ok.
  */
-int writePidFile(const TCHAR *filename, DWORD pid, int newUmask) {
+int writePidFile(const TCHAR *filename, DWORD pid, int newUmask, int strict) {
     FILE *pid_fp = NULL;
     int old_umask;
 
+    if (strict && wrapperFileExists(filename)) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("%d pid file, %s, already exists."), pid, filename);
+        return 1;
+    }
+
     old_umask = umask(newUmask);
     pid_fp = _tfopen(filename, TEXT("w"));
     umask(old_umask);
@@ -212,7 +225,7 @@ void takeSignalAction(int sigNum, const TCHAR *sigName, int mode) {
         switch (mode) {
         case WRAPPER_SIGNAL_MODE_RESTART:
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                TEXT("%s trapped.  Restarting JVM."), sigName);
+                TEXT("%s trapped.  %s"), sigName, wrapperGetRestartProcessMessage());
             wrapperRestartProcess();
             break;
 
@@ -228,13 +241,18 @@ void takeSignalAction(int sigNum, const TCHAR *sigName, int mode) {
                 (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH)) {
 
                 /* Signaled while we were already shutting down. */
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("%s trapped.  Forcing immediate shutdown."), sigName);
-
-                /* Disable the thread dump on exit feature if it is set because it
-                 *  should not be displayed when the user requested the immediate exit. */
-                wrapperData->requestThreadDumpOnFailedJVMExit = FALSE;
-                wrapperKillProcess();
+                if (wrapperData->isForcedShutdownDisabled) {
+                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
+                        TEXT("%s trapped.  Already shutting down."), sigName);
+                } else {
+                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
+                        TEXT("%s trapped.  Forcing immediate shutdown."), sigName);
+    
+                    /* Disable the thread dump on exit feature if it is set because it
+                     *  should not be displayed when the user requested the immediate exit. */
+                    wrapperData->requestThreadDumpOnFailedJVMExit = FALSE;
+                    wrapperKillProcess(FALSE);
+                }
             } else {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
                     TEXT("%s trapped.  Shutting down."), sigName);
@@ -257,7 +275,7 @@ void takeSignalAction(int sigNum, const TCHAR *sigName, int mode) {
             if (wrapperData->javaPID > 0) {
                 if (wrapperData->isDebugging) {
                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG,
-                        TEXT("%s trapped.  Forwarding to JVM process."), sigName);
+                        TEXT("%s (%d) trapped.  Forwarding to JVM process."), sigName, sigNum);
                 }
                 if (kill(wrapperData->javaPID, sigNum)) {
                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
@@ -269,6 +287,25 @@ void takeSignalAction(int sigNum, const TCHAR *sigName, int mode) {
             }
             break;
 
+        case WRAPPER_SIGNAL_MODE_PAUSE:
+            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
+                TEXT("%s trapped.  Pausing the application."), sigName);
+            wrapperPauseProcess(WRAPPER_ACTION_SOURCE_CODE_SIGNAL);
+            break;
+
+        case WRAPPER_SIGNAL_MODE_RESUME:
+            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
+                TEXT("%s trapped.  Resuming the application."), sigName);
+            wrapperResumeProcess(WRAPPER_ACTION_SOURCE_CODE_SIGNAL);
+            break;
+
+        case WRAPPER_SIGNAL_MODE_CLOSE_LOGFILE:
+            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
+                TEXT("%s trapped.  Closing the log file."), sigName);
+            flushLogfile();
+            closeLogfile();
+            break;
+
         default: /* WRAPPER_SIGNAL_MODE_IGNORE */
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
                 TEXT("%s trapped, but ignored."), sigName);
@@ -417,7 +454,7 @@ void descSignal(siginfo_t *sigInfo) {
  #ifdef UNICODE
     size_t req;
  #endif
-    TCHAR *uName;
+    TCHAR uName[MAX_USER_NAME_LENGTH + 1];
 #endif
 
     /* Not supported on all platforms */
@@ -451,23 +488,25 @@ void descSignal(siginfo_t *sigInfo) {
                 sigInfo->si_errno,
                 strerror(sigInfo->si_errno));
         }
-
+            
 #ifdef SI_USER
         if (sigInfo->si_code == SI_USER) {
             pw = getpwuid(sigInfo->si_uid);
             if (pw == NULL) {
-                uName = TEXT("<unknown>");
+                _sntprintf(uName, MAX_USER_NAME_LENGTH + 1, TEXT("<unknown>"));
             } else {
  #ifndef UNICODE
-                uName = pw->pw_name;
+                _sntprintf(uName, MAX_USER_NAME_LENGTH + 1, TEXT("%s"), pw->pw_name);
  #else
-                req = mbstowcs(NULL, pw->pw_name, 0) + 1;
-                uName = malloc(req * sizeof(TCHAR));
-                if (!uName) {
-                    outOfMemory(TEXT("DSCS"), 1);
+                req = mbstowcs(NULL, pw->pw_name, MBSTOWCS_QUERY_LENGTH);
+                if (req == (size_t)-1) {
                     return;
                 }
+                if (req > MAX_USER_NAME_LENGTH) {
+                    req = MAX_USER_NAME_LENGTH;
+                }
                 mbstowcs(uName, pw->pw_name, req + 1);
+                uName[req] = TEXT('\0'); /* Avoid bufferflows caused by badly encoded characters. */
  #endif
             }
 
@@ -479,9 +518,6 @@ void descSignal(siginfo_t *sigInfo) {
                 TEXT("  signal generated by PID: %d (Session PID: %d), UID: %d (%s)"),
  #endif
                 sigInfo->si_pid, getsid(sigInfo->si_pid), sigInfo->si_uid, uName);
- #ifdef UNICODE
-            free(uName);
- #endif
         }
 #endif
     }
@@ -556,7 +592,7 @@ void sigActionChildDeath(int sigNum, siginfo_t *sigInfo, void *na) {
     /* On UNIX, when a Child process changes state, a SIGCHLD signal is sent to the parent.
      *  The parent should do a wait to make sure the child is cleaned up and doesn't become
      *  a zombie process. */
-
+                
     threadId = pthread_self();
     if (timerThreadSet && pthread_equal(threadId, timerThreadId)) {
         if (wrapperData->isDebugging) {
@@ -687,8 +723,8 @@ void *javaIORunner(void *arg) {
     /* Loop until we are shutting down, but continue as long as there is more output from the JVM. */
     while ((!stopJavaIOThread) || (!nextSleep)) {
         if (nextSleep) {
-            /* Sleep for a hundredth of a second. */
-            wrapperSleep(10);
+            /* Sleep as little as possible. */
+            wrapperSleep(1);
         }
         nextSleep = TRUE;
         
@@ -697,7 +733,7 @@ void *javaIORunner(void *arg) {
             wrapperData->pauseThreadJavaIO = 0;
         }
         
-        if (wrapperReadChildOutput()) {
+        if (wrapperReadChildOutput(0)) {
             if (wrapperData->isDebugging) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG,
                     TEXT("Pause reading child process output to share cycles."));
@@ -754,7 +790,7 @@ void disposeJavaIO() {
 #endif
             wrapperSleep(100);
         }
-        pthread_kill(javaIOThreadId, SIGKILL);
+        pthread_cancel(javaIOThreadId);
     }
 }
 
@@ -803,6 +839,8 @@ void *timerRunner(void *arg) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Timer thread started."));
     }
 
+    wrapperGetSystemTicks();
+
     while (!stopTimerThread) {
         wrapperSleep(WRAPPER_TICK_MS);
         
@@ -838,15 +876,15 @@ void *timerRunner(void *arg) {
             first = FALSE;
         } else {
             if (offsetDiff > wrapperData->timerSlowThreshold) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_INFO,
+                log_printf_queue(TRUE, WRAPPER_SOURCE_WRAPPER, LEVEL_INFO,
                     TEXT("The timer fell behind the system clock by %ldms."), offsetDiff * WRAPPER_TICK_MS);
             } else if (offsetDiff < -1 * wrapperData->timerFastThreshold) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_INFO,
+                log_printf_queue(TRUE, WRAPPER_SOURCE_WRAPPER, LEVEL_INFO,
                     TEXT("The system clock fell behind the timer by %ldms."), -1 * offsetDiff * WRAPPER_TICK_MS);
             }
 
             if (wrapperData->isTickOutputEnabled) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT(
+                log_printf_queue(TRUE, WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT(
                     "    Timer: ticks=0x%08x, system ticks=0x%08x, offset=0x%08x, offsetDiff=0x%08x"),
                     nowTicks, sysTicks, tickOffset, offsetDiff);
             }
@@ -904,7 +942,7 @@ void disposeTimer() {
 #endif
             wrapperSleep(100);
         }
-        pthread_kill(timerThreadId, SIGKILL);
+        pthread_cancel(timerThreadId);
     }
 }
 
@@ -1051,6 +1089,17 @@ int wrapperBuildJavaCommand() {
     int length, i;
 
     /* If this is not the first time through, then dispose the old command array */
+    if (wrapperData->jvmVersionCommand) {
+        i = 0;
+        while(wrapperData->jvmVersionCommand[i] != NULL) {
+            free(wrapperData->jvmVersionCommand[i]);
+            wrapperData->jvmVersionCommand[i] = NULL;
+            i++;
+        }
+
+        free(wrapperData->jvmVersionCommand);
+        wrapperData->jvmVersionCommand = NULL;
+    }
     if (wrapperData->jvmCommand) {
         i = 0;
         while(wrapperData->jvmCommand[i] != NULL) {
@@ -1079,28 +1128,37 @@ int wrapperBuildJavaCommand() {
         return TRUE;
     }
 
-    if (wrapperData->commandLogLevel != LEVEL_NONE) {
-        for (i = 0; i < length; i++) {
-            log_printf(WRAPPER_SOURCE_WRAPPER, wrapperData->commandLogLevel,
-                TEXT("Command[%d] : %s"), i, strings[i]);
-        }
-
-        if (wrapperData->environmentClasspath) {
-            log_printf(WRAPPER_SOURCE_WRAPPER, wrapperData->commandLogLevel, TEXT(
-                "Classpath in Environment : %s"), wrapperData->classpath);
-        }
+    /* Allocate memory to hold array of version command strings.  The array is itself NULL terminated */
+    wrapperData->jvmVersionCommand = malloc(sizeof(TCHAR *) * (2 + 1));
+    if (!wrapperData->jvmVersionCommand) {
+        outOfMemory(TEXT("WBJC"), 1);
+        return TRUE;
     }
-
-    if (wrapperData->environmentClasspath) {
-        setEnv(TEXT("CLASSPATH"), wrapperData->classpath, ENV_SOURCE_WRAPPER);
+    memset(wrapperData->jvmVersionCommand, 0, sizeof(TCHAR *) * (2 + 1));
+    /* Java Command */
+    wrapperData->jvmVersionCommand[0] = malloc(sizeof(TCHAR) * (_tcslen(strings[0]) + 1));
+    if (!wrapperData->jvmVersionCommand[0]) {
+        outOfMemory(TEXT("WBJC"), 2);
+        return TRUE;
     }
-
+    _tcsncpy(wrapperData->jvmVersionCommand[0], strings[0], _tcslen(strings[0]) + 1);
+    /* -version */
+    wrapperData->jvmVersionCommand[1] = malloc(sizeof(TCHAR) * (8 + 1));
+    if (!wrapperData->jvmVersionCommand[1]) {
+        outOfMemory(TEXT("WBJC"), 3);
+        return TRUE;
+    }
+    _tcsncpy(wrapperData->jvmVersionCommand[1], TEXT("-version"), 8 + 1);
+    /* NULL */
+    wrapperData->jvmVersionCommand[2] = NULL;
+    
     /* Allocate memory to hold array of command strings.  The array is itself NULL terminated */
     wrapperData->jvmCommand = malloc(sizeof(TCHAR *) * (length + 1));
     if (!wrapperData->jvmCommand) {
         outOfMemory(TEXT("WBJC"), 1);
         return TRUE;
     }
+    memset(wrapperData->jvmCommand, 0, sizeof(TCHAR *) * (length + 1));
     /* number of arguments + 1 for a NULL pointer at the end */
     for (i = 0; i <= length; i++) {
         if (i < length) {
@@ -1110,6 +1168,8 @@ int wrapperBuildJavaCommand() {
                 return TRUE;
             }
             _tcsncpy(wrapperData->jvmCommand[i], strings[i], _tcslen(strings[i]) + 1);
+            
+            wrapperData->jvmCommand[i] = wrapperPostProcessCommandElement(wrapperData->jvmCommand[i]);
         } else {
             wrapperData->jvmCommand[i] = NULL;
         }
@@ -1122,16 +1182,154 @@ int wrapperBuildJavaCommand() {
 }
 
 /**
- * Launches a JVM process and stores it internally.
+ * Calculate the total length of the environment assuming that they are separated by spaces.
  */
-void wrapperExecute() {
+size_t wrapperCalculateEnvironmentLength() {
+    /* The compiler won't let us reverence environ directly in the for loop on OSX because it is actually a function. */
+    char **environment = environ;
+    size_t i;
+    size_t len;
+    size_t lenTotal;
+    
+    i = 0;
+    lenTotal = 0;
+    while (environment[i]) {
+        /* All we need is the length so we don't actually have to convert them. */
+        len = mbstowcs(NULL, environment[i], MBSTOWCS_QUERY_LENGTH);
+        if (len == (size_t)-1) {
+            /* Invalid string.  Skip. */
+        } else {
+            /* Add length of variable + null + pointer to next element */
+            lenTotal += len + 1 + sizeof(char *);
+        }
+        i++;
+    }
+    /* Null termination of the list. */
+    lenTotal += sizeof(char *) + sizeof(char *);
+    
+    return lenTotal;
+}
+
+/**
+ * Create a child process to print the Java version running the command:
+ *    /path/to/java -version
+ *  After printing the java version, the process is terminated.
+ * 
+ * In case the JVM is slow to start, it will time out after
+ * the number of seconds set in "wrapper.java.version.timeout".
+ * 
+ * Note: before the timeout is reached, the user can ctrl+c to stop the Wrapper.
+ */
+void launchChildProcessPrintJavaVersion() {
+    int blockTimeout;      /* max time (in ms) to wait for the child process to terminate */
+    int result;            /* result of waitpid */
+    int status;            /* status of child process */
+    pid_t procJavaVersion; /* pid of the child process */
+    
+    /* Create the child process! */
+    procJavaVersion = fork();
+                
+    if (procJavaVersion >= 0) {
+        /* fork successful */
+        
+        if (procJavaVersion == 0) {
+            /* in the child process */
+            TCHAR *javaVersionArgv[3];
+            javaVersionArgv[0] = wrapperData->jvmCommand[0];
+            javaVersionArgv[1] = TEXT("-version");
+            javaVersionArgv[2] = 0;
+            _texecvp(wrapperData->jvmCommand[0], javaVersionArgv);
+        } else {
+            /* in the parent process */
+            /* Note: on CentOS, in case we don't call waitpid(), then the child process becomes a zombie */
+            /* Note2: use _tprintf to log messages instead of log_printf. With log_printf, the columns are printed twice */
+            
+            /* If the user set the value to 0, then we will wait indefinitely. */
+            blockTimeout = getIntProperty(properties, TEXT("wrapper.java.version.timeout"), DEFAULT_JAVA_VERSION_TIMEOUT) * 1000;
+            
+            if (blockTimeout > 0) {
+                while (((result = waitpid(procJavaVersion, &status, WNOHANG)) == 0) && (blockTimeout > 0)) {
+#ifdef _DEBUG
+                    _tprintf(TEXT("Child process: Java version: waiting... result=%d waitpidStatLoc=%d blockTimeout=%d\n"), result, status, blockTimeout);
+#endif
+                    wrapperSleep(100);
+                    blockTimeout -= 100;
+                }
+            } else {
+                /* Wait indefinately. */
+                result = waitpid(procJavaVersion, &status, 0);
+            }
+            
+            if (result == 0) {
+                /* Timed out. */
+                _tprintf(TEXT("Child process: Java version: timed out\n"));
+                kill(procJavaVersion, SIGKILL);
+            } else if (result > 0) {
+                /* Process completed. */
+#ifdef _DEBUG
+                _tprintf(TEXT("Child process: Java version: successful\n"));
+#endif
+            } else {
+                /* Wait failed. */
+                _tprintf(TEXT("Child process: Java version: wait failed\n"));
+                kill(procJavaVersion, SIGKILL);
+            }
+        }
+    } else {
+        /* Fork failed. */
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Unable to spawn process to output Java version: %s"), getLastErrorText());
+    }
+}
+
+/**
+ * Launches a JVM process and stores it internally.
+ *
+ * @return TRUE if there were any problems.  When this happens the Wrapper will not try to restart.
+ */
+int wrapperExecute() {
+    int i;
     pid_t proc;
+    int execErrno;
+    size_t lenCmd;
+    size_t lenEnv;
 
     /* Create the pipe. */
     if (pipe(pipedes) < 0) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                    TEXT("Could not init pipe: %s"), getLastErrorText());
-        return;
+        return TRUE;
+    }
+    
+    /* Log the Java commands. */
+    
+    /* If the JVM version printout is requested then log its command line first. */
+    if (wrapperData->printJVMVersion) {
+        if (wrapperData->isDebugging) {
+            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG, TEXT("Java Command Line (Query Java Version):"));
+            for (i = 0; wrapperData->jvmVersionCommand[i] != NULL; i++) {
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG,
+                    TEXT("  Command[%d] : %s"), i, wrapperData->jvmVersionCommand[i]);
+            }
+        }
+    }
+    
+    /* Log ghe application java command line */
+    if (wrapperData->commandLogLevel != LEVEL_NONE) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, wrapperData->commandLogLevel, TEXT("Java Command Line:"));
+        for (i = 0; wrapperData->jvmCommand[i] != NULL; i++) {
+            log_printf(WRAPPER_SOURCE_WRAPPER, wrapperData->commandLogLevel,
+                TEXT("  Command[%d] : %s"), i, wrapperData->jvmCommand[i]);
+        }
+
+        if (wrapperData->environmentClasspath) {
+            log_printf(WRAPPER_SOURCE_WRAPPER, wrapperData->commandLogLevel,
+                TEXT("  Classpath in Environment : %s"), wrapperData->classpath);
+        }
+    }
+
+    /* Update the CLASSPATH in the environment if requested so the JVM can access it. */ 
+    if (wrapperData->environmentClasspath) {
+        setEnv(TEXT("CLASSPATH"), wrapperData->classpath, ENV_SOURCE_APPLICATION);
     }
 
     /* Make sure the log file is closed before the Java process is created.  Failure to do
@@ -1141,6 +1339,10 @@ void wrapperExecute() {
      *  threads do not reopen the log file as the new process is being created. */
     setLogfileAutoClose(TRUE);
     closeLogfile();
+        
+    /* Reset the log duration so we get new counts from the time the JVM is launched. */
+    resetDuration();
+    
     /* Fork off the child. */
     proc = fork();
 
@@ -1148,13 +1350,13 @@ void wrapperExecute() {
         /* Fork failed. */
 
         /* Restore the auto close flag. */
-        setLogfileAutoClose(wrapperData->logfileInactivityTimeout <= 0);
+        setLogfileAutoClose(wrapperData->logfileCloseTimeout == 0);
 
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
                    TEXT("Could not spawn JVM process: %s"), getLastErrorText());
 
         /* The pipedes array is global so do not close the pipes. */
-
+        return TRUE;
     } else {
         /* Reset the exit code when we launch a new JVM. */
         wrapperData->exitCode = 0;
@@ -1178,14 +1380,18 @@ void wrapperExecute() {
             if (dup2(pipedes[PIPE_WRITE_END], STDOUT_FILENO) < 0) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                     TEXT("%sUnable to set JVM's stdout: %s"), LOG_FORK_MARKER, getLastErrorText());
-                return;
+                /* This process needs to end. */
+                exit(wrapperData->errorExitCode);
+                return TRUE; /* Will not get here. */
             }
 
-            /* Send errors to the pipe by dupicating the pipe fd and setting the copy as the stderr fd. */
+            /* Send errors to the pipe by duplicating the pipe fd and setting the copy as the stderr fd. */
             if (dup2(pipedes[PIPE_WRITE_END], STDERR_FILENO) < 0) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                     TEXT("%sUnable to set JVM's stderr: %s"), LOG_FORK_MARKER, getLastErrorText());
-                return;
+                /* This process needs to end. */
+                exit(wrapperData->errorExitCode);
+                return TRUE; /* Will not get here. */
             }
             
             /* Close both ends of the pipe as we have already duplicated the Write end for our purposes. */
@@ -1194,13 +1400,35 @@ void wrapperExecute() {
             close(pipedes[PIPE_WRITE_END]);
             pipedes[PIPE_WRITE_END] = -1;
 
+            /* forking at this point, the child process has set all pipes already, so no
+               further assignments needed */
+            if (wrapperData->printJVMVersion) {
+                launchChildProcessPrintJavaVersion();
+            }
+            
             /* The pipedes array is global so do not close the pipes. */
             /* Child process: execute the JVM. */
             _texecvp(wrapperData->jvmCommand[0], wrapperData->jvmCommand);
+            execErrno = errno;
 
             /* We reached this point...meaning we were unable to start. */
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
-                TEXT("%sUnable to start JVM: %s (%d)"), LOG_FORK_MARKER, getLastErrorText(), errno);
+                TEXT("%sUnable to start JVM: %s (%d)"), LOG_FORK_MARKER, getLastErrorText(), execErrno);
+            if (execErrno == E2BIG) {
+                /* Command line too long. */
+                /* Calculate the total length of the command line. */
+                lenCmd = 0;
+                for (i = 0; wrapperData->jvmCommand[i] != NULL; i++) {
+                    lenCmd += _tcslen(wrapperData->jvmCommand[i]) + 1;
+                }
+                lenEnv = wrapperCalculateEnvironmentLength();
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("%s  The generated command line plus the environment was larger than the maximum allowed."), LOG_FORK_MARKER);
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("%s  The current length is %d bytes of which %d is the command line, and %d is the environment."), LOG_FORK_MARKER, lenCmd + lenEnv + 1, lenCmd, lenEnv); 
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("%s  It is not possible to calculate an exact maximum length as it depends on a number of factors for each system."), LOG_FORK_MARKER);
+
+                /* TODO: Figure out a way to inform the Wrapper not to restart and try again as repeatedly doing this is meaningless. */
+            }
+
 
             if (wrapperData->isAdviserEnabled) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE, TEXT("%s"), LOG_FORK_MARKER );
@@ -1224,9 +1452,10 @@ void wrapperExecute() {
             }
 
             /* This process needs to end. */
-            exit(1);
+            exit(wrapperData->errorExitCode);
+            return TRUE; /* Will not get here. */
         } else {
-            /* We are the parent side. */
+            /* We are the parent side and need to assume that at this point the JVM is up. */
             wrapperData->javaPID = proc;
             
             /* Close the write end as it is not used. */
@@ -1234,7 +1463,7 @@ void wrapperExecute() {
             pipedes[PIPE_WRITE_END] = -1;
 
             /* Restore the auto close flag. */
-            setLogfileAutoClose(wrapperData->logfileInactivityTimeout <= 0);
+            setLogfileAutoClose(wrapperData->logfileCloseTimeout == 0);			
 
             /* The pipedes array is global so do not close the pipes. */
 
@@ -1253,19 +1482,20 @@ void wrapperExecute() {
 
             /* If a java pid filename is specified then write the pid of the java process. */
             if (wrapperData->javaPidFilename) {
-                if (writePidFile(wrapperData->javaPidFilename, wrapperData->javaPID, wrapperData->javaPidFileUmask)) {
+                if (writePidFile(wrapperData->javaPidFilename, wrapperData->javaPID, wrapperData->javaPidFileUmask, FALSE)) {
                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
                         TEXT("Unable to write the Java PID file: %s"), wrapperData->javaPidFilename);
                 }
             }
 
-            /* If a java id filename is specified then write the pid of the java process. */
+            /* If a java id filename is specified then write the Id of the java process. */
             if (wrapperData->javaIdFilename) {
-                if (writePidFile(wrapperData->javaIdFilename, wrapperData->jvmRestarts, wrapperData->javaIdFileUmask)) {
+                if (writePidFile(wrapperData->javaIdFilename, wrapperData->jvmRestarts, wrapperData->javaIdFileUmask, FALSE)) {
                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
                         TEXT("Unable to write the Java Id file: %s"), wrapperData->javaIdFilename);
                 }
             }
+            return FALSE;
         }
     }
 }
@@ -1413,7 +1643,7 @@ int wrapperGetProcessStatus(TICKS nowTicks, int sigChild) {
             /* Error requesting the status. */
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Unable to request JVM process status: %s"), getLastErrorText());
         }
-        exitCode = 1;
+        exitCode = wrapperData->errorExitCode;
         res = WRAPPER_PROCESS_DOWN;
         wrapperJVMProcessExited(nowTicks, exitCode);
     } else {
@@ -1540,7 +1770,7 @@ void daemonize(int argc, TCHAR** argv) {
     if ((pid = fork()) < 0) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Could not spawn daemon process: %s"),
             getLastErrorText());
-        appExit(1, argc, argv);
+        appExit(wrapperData->errorExitCode, argc, argv);
     } else if (pid != 0) {
         /* Intermediate process is now running.  This is the original process, so exit. */
 
@@ -1558,7 +1788,7 @@ void daemonize(int argc, TCHAR** argv) {
     if (setsid() == -1) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("setsid() failed: %s"),
            getLastErrorText());
-        appExit(1, argc, argv);
+        appExit(wrapperData->errorExitCode, argc, argv);
     }
 
     signal(SIGHUP, SIG_IGN); /* don't let future opens allocate controlling terminals */
@@ -1590,7 +1820,7 @@ void daemonize(int argc, TCHAR** argv) {
     if ((pid = fork()) < 0) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Could not spawn daemon process: %s"),
             getLastErrorText());
-        appExit(1, argc, argv);
+        appExit(wrapperData->errorExitCode, argc, argv);
     } else if (pid != 0) {
         /* Daemon process is now running.  This is the intermediate process, so exit. */
         /* Call exit rather than appExit as we are only exiting this process. */
@@ -1598,7 +1828,7 @@ void daemonize(int argc, TCHAR** argv) {
     }
 
     /* Restore the auto close flag in the daemonized process. */
-    setLogfileAutoClose(wrapperData->logfileInactivityTimeout <= 0);
+    setLogfileAutoClose(wrapperData->logfileCloseTimeout == 0);
 }
 
 
@@ -1606,11 +1836,11 @@ void daemonize(int argc, TCHAR** argv) {
  * Sets the working directory to that of the current executable
  */
 int setWorkingDir(TCHAR *app) {
-    TCHAR szPath[PATH_MAX + 1];
+    TCHAR *szPath;
     TCHAR* pos;
 
     /* Get the full path and filename of this program */
-    if (_trealpath(app, szPath) == NULL) {
+    if ((szPath = findPathOf(app, TEXT("Wrapper binary"))) == NULL) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("Unable to get the path for '%s'-%s"),
             app, getLastErrorText());
         return 1;
@@ -1625,6 +1855,7 @@ int setWorkingDir(TCHAR *app) {
     pos = _tcsrchr(szPath, TEXT('/'));
     if (pos == NULL) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("Unable to extract path from: %s"), szPath);
+        free(szPath);
         return 1;
     } else {
         /* Clip the path at the position of the last backslash */
@@ -1632,12 +1863,13 @@ int setWorkingDir(TCHAR *app) {
     }
 
     /* Set a variable to the location of the binary. */
-    setEnv(TEXT("WRAPPER_BIN_DIR"), szPath, ENV_SOURCE_WRAPPER);
+    setEnv(TEXT("WRAPPER_BIN_DIR"), szPath, ENV_SOURCE_APPLICATION);
 
-    if (wrapperSetWorkingDir(szPath)) {
+    if (wrapperSetWorkingDir(szPath, TRUE)) {
+        free(szPath);
         return 1;
     }
-
+    free(szPath);
     return 0;
 }
 
@@ -1647,6 +1879,8 @@ int setWorkingDir(TCHAR *app) {
 #ifndef CUNIT
 #ifdef UNICODE
 int main(int argc, char **cargv) {
+    size_t req;
+    TCHAR **argv;
 #else
 int main(int argc, char **argv) {
 #endif
@@ -1654,33 +1888,18 @@ int main(int argc, char **argv) {
     int i;
 #endif
     TCHAR *retLocale;
-#ifdef UNICODE
-    size_t req;
-    TCHAR **argv;
-    
-    /* Create UNICODE versions of the argv array for internal use. */
-    argv = malloc(argc * sizeof *argv );
-    if(!argv) {
-        _tprintf(TEXT("Out of Memory in Main\n"));
-        appExit(1, 0, NULL);
+
+#ifdef FREEBSD
+    /* In the case of FreeBSD, we need to dynamically load and initialize the iconv library to work with all versions of FreeBSD. */
+    if (loadIconvLibrary()) {
+        /* Already reported. */
+        /* Don't call appExit here as we are not far enough along. */
         return 1;
     }
-    for (i = 0; i < argc; i++) {
-        req = mbstowcs(NULL, cargv[i], 0);
-        argv[i] = malloc((int)(req + 1) * sizeof(TCHAR));
-        if (!argv[i]) {
-            _tprintf(TEXT("Out of Memory in Main\n"));
-            while(--i > 0) {
-                free(argv[i]);
-            }
-            free(argv);
-            appExit(1, 0, argv);
-            return 1;
-        }
-        mbstowcs(argv[i], cargv[i], (req + 1) * sizeof(TCHAR));
-    }
-
-#endif
+#endif  
+  
+    /* Set the default locale here so any startup error messages will have a chance of working.
+     *  This should be done before converting cargv to argv, because there might be accentued letters in cargv. */
     retLocale = _tsetlocale(LC_ALL, TEXT(""));
     if (retLocale) {
 #if defined(UNICODE)
@@ -1696,6 +1915,38 @@ int main(int argc, char **argv) {
 #endif
         */
     }
+    
+#ifdef UNICODE
+    /* Create UNICODE versions of the argv array for internal use. */
+    argv = malloc(argc * sizeof(TCHAR *));
+    if (!argv) {
+        _tprintf(TEXT("Out of Memory in Main\n"));
+        appExit(1, 0, NULL);
+        return 1;
+    }
+    for (i = 0; i < argc; i++) {
+        req = mbstowcs(NULL, cargv[i], MBSTOWCS_QUERY_LENGTH);
+        if (req == (size_t)-1) {
+            _tprintf(TEXT("Encoding problem with arguments in Main\n"));
+            free(argv);
+            appExit(1, 0, NULL);
+            return 1;
+        }
+        argv[i] = malloc(sizeof(TCHAR) * (req + 1));
+        if (!argv[i]) {
+            _tprintf(TEXT("Out of Memory in Main\n"));
+            while (--i > 0) {
+                free(argv[i]);
+            }
+            free(argv);
+            appExit(1, 0, NULL);
+            return 1;
+        }
+        mbstowcs(argv[i], cargv[i], req + 1);
+        argv[i][req] = TEXT('\0'); /* Avoid bufferflows caused by badly encoded characters. */
+    }
+#endif
+
     if (wrapperInitialize()) {
         appExit(1, argc, argv);
         return 1; /* For compiler. */
@@ -1754,6 +2005,12 @@ int main(int argc, char **argv) {
     }
 
     /* Load the properties. */
+    /* To make the WRAPPER_LANG references in the configuration work correctly,
+     *  it is necessary to load the configuration twice.
+     * The first time, we want to ignore the return value.  Any errors will be
+     *  suppressed and will get reported again the second time through. */
+    /* From version 3.5.27, the community edition will also preload the configuration properties. */ 
+    wrapperLoadConfigurationProperties(TRUE);
     if (wrapperLoadConfigurationProperties(FALSE)) {
         /* Unable to load the configuration.  Any errors will have already
          *  been reported. */
@@ -1762,7 +2019,8 @@ int main(int argc, char **argv) {
              *  it did not exist.  Show the usage. */
             wrapperUsage(argv[0]);
         }
-        appExit(1, argc, argv);
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("  The Wrapper will stop."));
+        appExit(wrapperData->errorExitCode, argc, argv);
         return 1; /* For compiler. */
     }
 
@@ -1780,6 +2038,9 @@ int main(int argc, char **argv) {
         /* fork to a Daemonized process if configured to do so. */
         if (wrapperData->daemonize) {
             daemonize(argc, argv);
+            
+            /* We are now daemonized, so mark this as being a service. */
+            wrapperData->isConsole = FALSE;
 
             /* When we daemonize the Wrapper, its PID changes. Because of the
              *  WRAPPER_PID environment variable, we need to set it again here
@@ -1797,9 +2058,10 @@ int main(int argc, char **argv) {
              *  the configuration can be reloaded.  This is needed to support relative
              *  references to include files. */
             if (wrapperData->workingDir && wrapperData->originalWorkingDir) {
-                if (wrapperSetWorkingDir(wrapperData->originalWorkingDir)) {
+                if (wrapperSetWorkingDir(wrapperData->originalWorkingDir, TRUE)) {
                     /* Failed to restore the working dir.  Shutdown the Wrapper */
-                    appExit(1, argc, argv);
+                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("  The Wrapper will stop."));
+                    appExit(wrapperData->errorExitCode, argc, argv);
                     return 1; /* For compiler. */
                 }
             }
@@ -1813,7 +2075,8 @@ int main(int argc, char **argv) {
                      *  it did not exist.  Show the usage. */
                     wrapperUsage(argv[0]);
                 }
-                appExit(1, argc, argv);
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("  The Wrapper will stop."));
+                appExit(wrapperData->errorExitCode, argc, argv);
                 return 1; /* For compiler. */
             }
         }
@@ -1825,28 +2088,42 @@ int main(int argc, char **argv) {
             rollLogs();
         }
 
+        if (wrapperData->pidFilename) {
+            if (writePidFile(wrapperData->pidFilename, wrapperData->wrapperPID, wrapperData->pidFileUmask, wrapperData->pidFileStrict)) {
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
+                     TEXT("ERROR: Could not write pid file %s: %s"),
+                wrapperData->pidFilename, getLastErrorText());
+                     /* Common wrapper cleanup code. */
+                wrapperDispose();
+#if defined(UNICODE)
+                for (i = 0; i < argc; i++) {
+                    if (argv[i]) {
+                        free(argv[i]);
+                    }
+                }
+                if (argv) {
+                    free(argv);
+                }
+#endif
+                exit(wrapperData->errorExitCode);
+                return 1; /* For compiler. */
+            }
+        }
+
         /* Write pid and anchor files as requested.  If they are the same file the file is
          *  simply overwritten. */
         if (wrapperData->anchorFilename) {
-            if (writePidFile(wrapperData->anchorFilename, wrapperData->wrapperPID, wrapperData->anchorFileUmask)) {
+            if (writePidFile(wrapperData->anchorFilename, wrapperData->wrapperPID, wrapperData->anchorFileUmask, FALSE)) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
                      TEXT("ERROR: Could not write anchor file %s: %s"),
                      wrapperData->anchorFilename, getLastErrorText());
-                appExit(1, argc, argv);
-                return 1; /* For compiler. */
-            }
-        }
-        if (wrapperData->pidFilename) {
-            if (writePidFile(wrapperData->pidFilename, wrapperData->wrapperPID, wrapperData->pidFileUmask)) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
-                     TEXT("ERROR: Could not write pid file %s: %s"),
-                     wrapperData->pidFilename, getLastErrorText());
-                appExit(1, argc, argv);
+                appExit(wrapperData->errorExitCode, argc, argv);
                 return 1; /* For compiler. */
             }
         }
+
         if (wrapperData->lockFilename) {
-            if (writePidFile(wrapperData->lockFilename, wrapperData->wrapperPID, wrapperData->lockFileUmask)) {
+            if (writePidFile(wrapperData->lockFilename, wrapperData->wrapperPID, wrapperData->lockFileUmask, FALSE)) {
                 /* This will fail if the user is running as a user without full privileges.
                  *  To make things easier for user configuration, this is ignored if sufficient
                  *  privileges do not exist. */
@@ -1857,13 +2134,17 @@ int main(int argc, char **argv) {
             }
         }
 
-        appExit(wrapperRunConsole(), argc, argv);
+        if (wrapperData->isConsole) {
+            appExit(wrapperRunConsole(), argc, argv);
+        } else {
+            appExit(wrapperRunService(), argc, argv);
+        }
         return 0; /* For compiler. */
     } else {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT(""));
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Unrecognized option: -%s"), wrapperData->argCommand);
         wrapperUsage(argv[0]);
-        appExit(1, argc, argv);
+        appExit(wrapperData->errorExitCode, argc, argv);
         return 1; /* For compiler. */
     }
 }
